// measuring internal temperature,
// hand-measured VDDA=VREF+

#include "stm32f4xx.h"
// #include "stm32f4xx_augment.h"
#define ADC_SMP__144     6      // 110: 144 cycles
#define ADC_CCR_ADCPRE__PCLK2_2  0   /* PCLK2/4 */

// #include "common.h"
// #include "common_gcc.h"


//---------------------------- app defines
#define USE_VREFINT 1
#define USE_SLEEP   0  // requires to break in debugger to read out tempint


// default HSI 16MHz clock -- but we want to divide it down
#define CLOCK_MHZ  2   // can use 1, 2, 4, 8 and 16 -- see InitClock() -- note that min. ADC clock is 600kHz and ADC prescaler is min. 2, so 1MHz sysclk is out of spec for ADC


/*
Disco F407
4LEDS on PD12..PD15, these are also TIM4_CH1..CH4
*/
#define LED_PORT     GPIOD
#define LED_PIN      15

// BUTTON     PA0, switches to VCC
#define BUTTON_PORT      GPIOA
#define BUTTON_PIN       0


// -------- utils/frameworks
// trivial loopdelay
static void LoopDelay(volatile uint32_t n) {
	while(n > 0) n--;
}

#define DELAY_CONSTANT (CLOCK_MHZ * 300000)

// ---------- clock

// default HSI clock, divided down
static void InitClock(void) {
  uint32_t hpre;

  switch(CLOCK_MHZ) {
    default:
    case 16:
      return;
    case 8:
      hpre = 0b1000; break;
    case 4:
      hpre = 0b1001; break;
    case 2:
      hpre = 0b1010; break;
    case 1:
      hpre = 0b1011; break;
  }
  RCC->CFGR = (RCC->CFGR & ~RCC_CFGR_HPRE_Msk) | (hpre << RCC_CFGR_HPRE_Pos);
}


// -------- timer ISR for sleep
#if (USE_SLEEP)
  void TIM4_IRQHandler(void) __attribute__((interrupt));
  void TIM4_IRQHandler(void) {
    uint32_t sr;

    sr = TIM4->SR;
    TIM4->SR = ~sr;  // clear flags
    if (sr & TIM_SR_UIF) {
      // do nothing here, we just wanted to wake up
    }
  }
#endif
// --------- ADC

static void InitAdc(void) {
  // ADC->CCR = ADC_CCR_ADCPRE__PCLK2_2 << ADC_CCR_ADCPRE_Pos;  // ADC prescaler - minimum 2  -- this is default, no need to set it

  // sampling times shall be for internal sources mostly cca 10us, at SYSCLK max. 16MHz we have ADC clock 8MHz i.e. 125ns/tADC
  ADC1->SMPR1 = 0
    | ( ADC_SMP__144 * ADC_SMPR1_SMP16_0)  // tempsens // not used in 'F427, but should make no harm
    | ( ADC_SMP__144 * ADC_SMPR1_SMP17_0)  // vrefint
    | ( ADC_SMP__144 * ADC_SMPR1_SMP18_0)  // tempsens
  ;
  // enable voltage reference
  ADC->CCR |= ADC_CCR_TSVREFE;
  // start up the ADC:
  ADC1->CR2 = (1 * ADC_CR2_ADON);
}

#if (USE_VREFINT)
  volatile float vrefplus;
#else
  #define VREFPLUS 2.91f  // measured on R23=VREF+
#endif

enum {
  ADC_STATE__INIT = 0,
  // any ADC startup states go here
  ADC_STATE__FIRST,
  // now the individual runtime states
  ADC_STATE__TEMP = ADC_STATE__FIRST,
  #if (USE_VREFINT)
    ADC_STATE__VREFINT,
  #endif
  // ADC_STATE__xxx
  ADC_STATE__END,
};

uint16_t adcRaw[ADC_STATE__END - ADC_STATE__FIRST];
volatile float tempint;

static void HandleAdc(void) {
  static uint8_t state = ADC_STATE__INIT;
  uint16_t val;

  if (state == ADC_STATE__INIT) {
    // perform whatever needed to init ADC - calibration for example - that's not needed for the 'F4 so there's nothing here
    state = ADC_STATE__FIRST;
  } else {  // runtime acquisition in progress
    if ((ADC1->SR & ADC_SR_EOC) == 0) return;
    val = ADC1->DR;
    switch (state) {
      case ADC_STATE__TEMP:
        adcRaw[ADC_STATE__TEMP - ADC_STATE__FIRST] = val;
        // ADC->CCR &= ~ADC_CCR_TSVREFE;
        break;
      #if (USE_VREFINT)
        case ADC_STATE__VREFINT:
          adcRaw[ADC_STATE__VREFINT - ADC_STATE__FIRST] = val;
          // ADC->CCR &= ~ADC_CCR_TSVREFE;
          break;
      #endif
    }
    state++;
    if (state == ADC_STATE__END) {   // end of conversions cycle, perform calculations
      {
        float temp;

        temp = (float)adcRaw[ADC_STATE__TEMP - ADC_STATE__FIRST];
        #if (USE_VREFINT)  // use indirect reference derived from VREFINT measurement
          temp = temp * (float)*(uint16_t *)0x1FFF7A2A / (float)adcRaw[ADC_STATE__VREFINT - ADC_STATE__FIRST]; // compensate difference of stored and measured vref (i.e. difference between the supply voltages used as ADC reference)
        #else              // use explicitly measured VREF+
          temp = temp * VREFPLUS / 3.3f; // factory calibration TS_CALx was at 3.3V
        #endif
        // now interpolate using the stored values for 30 and 110 deg.
        temp = (temp - (float)*(uint16_t *)0x1FFF7A2C) / ((float)*(uint16_t *)0x1FFF7A2E - (float)*(uint16_t *)0x1FFF7A2C);
        temp = temp * (110.0f - 30.0f) + 30.0f;

        tempint = temp;

        #if (USE_VREFINT)
          vrefplus = 3.3f * (float)*(uint16_t *)0x1FFF7A2A / (float)adcRaw[ADC_STATE__VREFINT - ADC_STATE__FIRST];  // 0x1FFF7A2A - VREFIN_CAL Raw data acquired at temperature of 30 C, VDDA = 3.3 V
        #endif
      }
      state = ADC_STATE__FIRST; // and restart
    }
  }
  // go for the next acq
  switch (state) {
    case ADC_STATE__TEMP:
      // ADC->CCR |= ADC_CCR_TSVREFE;
      ADC1->SQR3 = (16 << ADC_SQR3_SQ1_Pos);
      break;
    #if (USE_VREFINT)
      case ADC_STATE__VREFINT:
        // ADC->CCR |= ADC_CCR_TSVREFE;
        ADC1->SQR3 = (17 << ADC_SQR3_SQ1_Pos);
        break;
    #endif
  }
  // we might perhaps want to set also the number of conversions, which is in SQ1 register, but it's by default 0 which is 1 conversion which is OK for us.
  // start one conversion
  ADC1->CR2 |= (uint32_t)ADC_CR2_SWSTART;
}


// ---------------------------- MAIN ------------------------------
int main(void) {

  InitClock();

  RCC->AHB1ENR |= 0
    | RCC_AHB1ENR_GPIODEN
  ;

  RCC->APB1ENR |= 0
    | RCC_APB1ENR_TIM4EN
  ;
  RCC->APB2ENR |= 0
    | RCC_APB2ENR_ADC1EN
  ;


// in GPIOA->MODER, PA0 is at default as Input
// and there's also an external pulldown, so no need to switch on the internal one


  GPIOD->MODER = (GPIOD->MODER  // assuming reset value
  ) | (0
    | (0b10 << GPIO_MODER_MODER15_Pos)   // blue LED, 0b10 = GPIO_Mode_AF
  );
  GPIOD->AFR[1] = GPIOD->AFR[1] | (0
    | (2 << GPIO_AFRH_AFSEL15_Pos)  // TIM4 is AF2
  );

  TIM4->CCER = 0
    | TIM_CCER_CC4E // enable TIM4_CH4
  ;
  #if (USE_SLEEP)
    TIM4->PSC = (CLOCK_MHZ * 1000) - 1;  // so that ARR ticks at 1kHz - PSC is preloaded so first period will be shorted but we don't really care here
    TIM4->ARR = 1000 - 1;                // overflow in each 1s
    TIM4->DIER = TIM_DIER_UIE;           // enable interrupt upon overflow = Update
    TIM4->CR1 = TIM_CR1_CEN;             // enable counter
    NVIC_EnableIRQ(TIM4_IRQn);
  #endif

  InitAdc();

  while(1) {
    HandleAdc();
    TIM4->CCMR2 = 0
      | (0b00  << TIM_CCMR2_CC4S_Pos)  // CH4 select 0b00 = output compare
      | (0b101 << TIM_CCMR2_OC4M_Pos)  // 0b101 = Force Active
    ;
    #if (USE_SLEEP)
      __WFI();
    #else
      LoopDelay(DELAY_CONSTANT);
      if (BUTTON_PORT->IDR & (1 << BUTTON_PIN)) {
        LoopDelay(3 * DELAY_CONSTANT);
      }
    #endif
    TIM4->CCMR2 = 0
      | (0b00  << TIM_CCMR2_CC4S_Pos)  // CH4 select 0b00 = output compare
      | (0b100 << TIM_CCMR2_OC4M_Pos)  // 0b100 = Force Inactive
    ;
    #if (USE_SLEEP)
      __WFI();
    #else
      LoopDelay(DELAY_CONSTANT);
    #endif
  }
}
